#mstr-modal
  .modal-footer {
    padding: 0px;
    display: none;
}

#mstr-modal.modal.fade
  .modal-dialog {
    max-height: 661px;
    max-width: 538px;
    height: 90%;
}

#mstr-modal
  .modal-dialog {
    margin: 30px auto;
    height: 90%;
}

#mstr-modal
  .modal-header {
    padding: 0px;
    border-bottom: none;
    position: absolute;
    top: -1px;
    right: 2px;
    float: right;
    z-index: 1;
}

#mstr-modal
  .modal-header
    .close {
      margin-top: unset;
}

#mstr-modal
  .modal-header > button {
    padding: 3px 8px 6px 8px;
}

#mstr-modal
  .modal-body {
    padding: 0px 0px 0px 0px;
    height: 100%;
}

#mstr-modal
  .top-nav {
    padding-bottom: 10px;
}

#mstr-modal
  .header {
    height: 50px;
}

#mstr-modal
  .modal-content {
    height: 100%;
}

#mstr-modal
  .modal-body > div {
    height: inherit;
}

#mstr-modal * {
  box-sizing: border-box;
}

#mstr-ico {
  padding-bottom: 2px;
}

@media (max-height: 560px) {
  #mstr-modal
    .modal.fade
      .modal-dialog {
        top: -15px;
  }
}

#mstr-copy-to-clipboard-object {
  position: absolute;
  top: 0;
  left: 0;
  pointer-events: none;
  width: 0.1px;
  height: 0.1px;
}

/* IFRAME */
#mstr-modal iframe {
  display: block;
  width: 100%;
  height: 100%;
  border: none;
  padding-top: 5px;
}


/* Custom Cell Engine */
.force-hidden {
  display: none !important;
}

.force-disabled {
  pointer-events: none;
  opacity: 0.5;
}

.input_area.visible {
  position: relative;
  z-index: 1;
}

.input_area.visible
  .CodeMirror-lines {
    cursor: default;
}

.input_area.visible
  .mstr-buttons-container {
    width: 0;
    position: relative;
    z-index: 3;
    margin: 8px 0;
    display: flex;
    flex-wrap: nowrap;
    justify-content: flex-end;
    float: right;
}

.input_area.visible
  .mstr-buttons-container:after { /* floating margin, so the code cell is not affected */
    content: '';
    margin-right: 8px;
  }

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button {
      cursor: pointer;
      border: 0;
      border-radius: 3px;
      margin-left: 8px;
      padding: 0 5px;
      box-shadow: 0 0.5px 3px 0 rgba(0, 0, 0, 0.3);
      background: rgb(250, 250, 250);
      min-width: 34px;
      min-height: 28px;
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button:hover {
      box-shadow:
        1px 1px 3px rgba(100, 100, 100, 0.2),
        -1px -1px 3px rgba(100, 100, 100, 0.2),
        -1px 1px 3px rgba(100, 100, 100, 0.2),
        1px -1px 3px rgba(100, 100, 100, 0.2);
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button:focus {
      outline: 0;
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button
      img {
        max-width: 18px;
        max-height: 22px;
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button
      .mstr-tooltip {
        position: absolute;
        background: #404040;
        color: #fff;
        top: calc(100% + 16px);
        padding: 12px;
        border-radius: 3px;
        font-size: 12px;
        transform: translateX(-100%) translateX(35px);
        display: flex;
        pointer-events: none;
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button
      .mstr-tooltip
        span {
          position: relative;
          z-index: 110;
          white-space: pre;
}

.input_area.visible
  .mstr-buttons-container
    .mstr-cell-button
      .mstr-tooltip
        .arrow {
          background: #404040;
          width: 16px;
          height: 16px;
          position: absolute;
          right: 15px;
          top: -6px;
          transform: rotate(45deg);
          z-index: 109;
}
